import { AcceptedZipSources, BaseArchive } from './StreamArchive';
import { ArchiveIndex, PartialTweetGDPR, PartialTweet, AccountGDPR, ProfileGDPR, ClassicPayloadDetails, TwitterUserDetails, DMFile, PartialTweetUser, InnerGDPRPersonalization, GPDRScreenNameHistory, GPDRProtectedHistory, InnerGDPRAgeInfo, GDPRMoment, DirectMessage } from './TwitterTypes';
import DMArchive from './DMArchive';
import { EventTarget } from 'event-target-shim';
export declare type ArchiveReadState = "idle" | "reading" | "indexing" | "tweet_read" | "user_read" | "dm_read" | "extended_read" | "ready";
/** Raw informations stored in GDPR, extracted for a simpler use.
 *
 * This includes list of followers, followings, favorites, mutes, blocks,
 * registered and subscribed lists, history of screen names, and Twitter moments.
 */
export interface ExtendedGDPRInfo {
    followers: Set<string>;
    followings: Set<string>;
    favorites: Set<string>;
    mutes: Set<string>;
    blocks: Set<string>;
    lists: {
        created: string[];
        member_of: string[];
        subscribed: string[];
    };
    personalization: InnerGDPRPersonalization;
    screen_name_history: GPDRScreenNameHistory[];
    protected_history: GPDRProtectedHistory[];
    age_info: InnerGDPRAgeInfo;
    moments: GDPRMoment[];
}
/** Return the `Date` object affiliated to **tweet**. */
export declare function dateFromTweet(tweet: PartialTweet): Date;
export declare function parseTwitterDate(date: string): Date;
/**
 * Return true if **tweet** contains media(s).
 *
 * This includes photos, videos or animated GIF.
 */
export declare function isWithMedia(tweet: PartialTweet): boolean;
/**
 * Return true if **tweet** contains a video or one animated GIF.
 *
 * Twitter's GIF are mp4 encoded.
 */
export declare function isWithVideo(tweet: PartialTweet): boolean;
declare type TwitterArchiveEvents = {
    zipready: CustomEvent<void>;
    userinfosready: CustomEvent<void>;
    tweetsread: CustomEvent<void>;
    indexready: CustomEvent<void>;
    willreaddm: CustomEvent<void>;
    willreadextended: CustomEvent<void>;
    ready: CustomEvent<void>;
    error: CustomEvent<any>;
};
declare type TwitterArchiveOnEvents = {
    onzipready: CustomEvent<void>;
    onuserinfosready: CustomEvent<void>;
    ontweetsread: CustomEvent<void>;
    onindexready: CustomEvent<void>;
    onwillreaddm: CustomEvent<void>;
    onwillreadextended: CustomEvent<void>;
    onready: CustomEvent<void>;
    onerror: CustomEvent<any>;
};
/**
 * Represents a full TwitterArchive. Support GDPR and classic archive.
 *
 * Remember that tweets, in searchs in particular are **NOT** sorted.
 *
 * Direct messages, parsed if archive is a GDPR archive, stored in `.messages`,
 * are returned and sorted from the most older to the more recent.
 */
export declare class TwitterArchive extends EventTarget<TwitterArchiveEvents, TwitterArchiveOnEvents> {
    protected load_images_in_zip: boolean;
    protected _ready: Promise<void>;
    protected archive: BaseArchive<any>;
    /** Current archive load state. */
    state: ArchiveReadState;
    /**
     * Access to extended infos. See `ExtendedGDPRInfo` interface.
     *
     * Defined only if `.is_gdpr === true`.
     * */
    extended_gdpr: ExtendedGDPRInfo;
    protected _index: ArchiveIndex;
    protected dms: DMArchive;
    protected dm_img_archive: BaseArchive<any>;
    protected dm_img_group_archive: BaseArchive<any>;
    protected _is_gdpr: boolean;
    protected user_cache: PartialTweetUser;
    /**
     * Build a new instance of TwitterArchive.
     *
     * @param file Accept any source that `JSZip` library accepts, and string for filenames.
     * If this parameter is `null`, then you'll need to load each part one by one !
     * @param build_extended True if `.extended_gdpr` should be built (only if **file** is a GDPR archive.)
     * @param keep_loaded If possible, free the memory after load if set to false.
     * @param load_images_in_zip In Twitter GDPR archives v2, tweet and dm images are in ZIP archives inside the ZIP.
     * If `true`, TwitterArchive will extract its content in RAM to allow the usage of images.
     * If `false`, DMs images will be unavailable.
     * If `undefined`, Twitter will extract in RAM in browser mode, and leave the ZIP untouched in Node.js.
     *
     * If you want to save memory, set this parameter to `false`,
     * and before using `.dmImage()` methods, check if you need to load DM images ZIP
     * with `.requiresDmImageZipLoad()`.
     *
     * Then, if you need to, load the DM image ZIP present in the archive using `.loadCurrentDmImageZip()`.
     * **Please note that `keep_loaded` should be set to `true` to use this method !**
     */
    constructor(file: AcceptedZipSources | Promise<AcceptedZipSources> | null, build_extended?: boolean, keep_loaded?: boolean, load_images_in_zip?: boolean);
    protected initGDPR(extended: boolean, keep_loaded: boolean): Promise<void>;
    protected initExtendedGDPR(): Promise<void>;
    protected convertToPartial(tweet: PartialTweetGDPR, pic_prof: string): PartialTweet;
    protected initClassic(): Promise<void>;
    protected isGDPRArchive(): boolean;
    /**
     * True if given tweet is coming from a GDPR archive.
     *
     * Tweets getted by available getters are NOT GDPR tweets, they've been converted !
     */
    isGDPRTweet(tweet: PartialTweetGDPR | PartialTweet): tweet is PartialTweetGDPR;
    /** Extract tweets from a specific month. */
    month(month: string, year: string): PartialTweet[];
    /** Find tweets made on the same day (= month, = day), but in all years. */
    fromThatDay(): PartialTweet[];
    /** Get tweets in a specific time interval. */
    between(since: Date, until: Date): PartialTweet[];
    /** Get a single tweet by ID. Returns `null` if tweet does not exists. */
    id(id_str: string): PartialTweet | null;
    /**
     * Give the media url in direct message, obtain the Blob-bed image.
     * For use in Node.js, you must set `as_array_buffer` to `true` !
     */
    dmImageFromUrl(url: string, is_group?: boolean, as_array_buffer?: boolean): Promise<ArrayBuffer | Blob>;
    /**
     * Extract a direct message image from GDPR archive (exact filename required).
     * For use in Node.js, you must set `as_array_buffer` to `true` !
     */
    dmImage(name: string, is_group?: boolean, as_array_buffer?: boolean): Promise<Blob | ArrayBuffer>;
    /**
     * Return all the images of a direct message, as blob or array buffer.
     *
     * If the message does not exists or the DM archive is not loaded / available,
     * return an empty array.
     *
     * Otherwise, return a array of `Blob` / `ArrayBuffer`
     *
     * @param direct_message Direct Message id | Direct message
     * @param as_array_buffer Return an `ArrayBuffer` array, instead of a `Blob` array
     */
    dmImagesOf(direct_message: string | DirectMessage, as_array_buffer?: boolean): Promise<(Blob | ArrayBuffer)[]>;
    /**
     * Return true if you need to load a DM image ZIP in order to use `.dmImage()`.
     *
     * If you need to, see `.loadCurrentDmImageZip()`, or `importDmImageZip()`.
     */
    requiresDmImageZipLoad(): boolean;
    /**
     * Load/reload zip archives that contains DM images,
     * if `load_images_in_zip` parameter was set to false.
     *
     * You need to have the archive loaded to accomplish this action
     * (constructor `keep_loaded` parameter should be set to `true`).
     *
     * Note that if any zip is found in the archive, this method will just do nothing.
     */
    loadCurrentDmImageZip(): Promise<void>;
    /**
     * Import a custom ZIP file as DM single-conversation images file.
     */
    importDmImageZip(file: AcceptedZipSources | Promise<AcceptedZipSources>): Promise<void>;
    /**
     * Import a custom ZIP file as DM group-conversation images file.
     */
    importDmGroupImageZip(file: AcceptedZipSources | Promise<AcceptedZipSources>): Promise<void>;
    /** All tweets registered in this archive. */
    get all(): PartialTweet[];
    /** Access to the DMArchive object. Will be undefined if `.is_gdpr === false`. */
    get messages(): DMArchive;
    /** ID of the user who created this archive. */
    get owner(): string;
    /**
     * Screen name (@) of the user who created this archive.
     * May be obsolete (user can change screen_name over time).
     */
    get owner_screen_name(): string;
    /** Archive creation date. Not accurate in GDPR archive (will be the current date). */
    get generation_date(): Date;
    /** Archive information and tweet index. */
    get index(): ArchiveIndex;
    /** Number of tweets in this archive. */
    get length(): number;
    /** True if archive is a GDPR archive. */
    get is_gdpr(): boolean;
    /** Raw Archive object. Can be used to get specific files. */
    get raw(): BaseArchive<any>;
    protected get should_autoload_zip_img(): boolean;
    /** Resolved when archive read is over. */
    ready(): Promise<void>;
    /** -------------------- */
    /** SIDELOADING MANUALLY */
    /** -------------------- */
    /**
     * Load a part of a GDPR archive.
     *
     * Set current archive as GDPR archive.
     */
    loadArchivePart(parts?: {
        tweets?: PartialTweetGDPR[];
        account?: AccountGDPR;
        profile?: ProfileGDPR;
        dms?: DMFile[];
    }): void;
    /**
     * Load a part of a classic archive.
     *
     * Set current archive as non-gdpr archive.
     */
    loadClassicArchivePart(parts?: {
        tweets?: PartialTweet[];
        payload?: ClassicPayloadDetails;
        user?: TwitterUserDetails;
    }): void;
    /**
     * Convert GDPR tweets into classic tweets, then register them into the index.
     */
    protected readGDPRTweets(tweets: PartialTweetGDPR[], profile_img_https: string): void;
    /**
     * Register tweets into the index.
     */
    protected readTweets(tweets: PartialTweet[]): void;
    /**
     * Sort tweets by ID.
     */
    protected sortTweets(tweets: PartialTweet[]): PartialTweet[];
}
export {};
